/*
     This file is part of vgp daemon.
     (C) 2003 Marco Tizzoni aka elibus (elibus@freaknet.org)
     Key available on key server.
     Key fingerprint: 3A97 6C78 0D07 91FC 0509  0661 DAAE F7B4 FC6F E313

     vgp daemon is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     vgpd is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with libextractor; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <features.h>		/* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>	/* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>	/* The L2 protocols */
#endif


#include <sys/ioctl.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>

#include <linux/route.h>

#include "vgp.h"
#include "if_util.h"
 extern int debug;
				/*
				 * retrieve ethernet interface index
				 */ int
get_ifindex (int *s, char *ifname)
{
  struct ifreq ifr;
  memset (&ifr, 0, sizeof (struct ifreq));
  strncpy (ifr.ifr_name, ifname, IFNAMSIZ);
  if (debug)
    syslog (LOG_DEBUG, "ifname: %s\n", ifname);
  if (ioctl (*s, SIOCGIFINDEX, &ifr) == -1)
    {
      syslog (LOG_ERR, "get_ifindex(%s): SIOCGIFINDEX: %s", ifname,
	      strerror (errno));
      return (-1);
    }
  if (debug)
    syslog (LOG_DEBUG, "Successfully got interface index: %i\n",
	    ifr.ifr_ifindex);
  return ifr.ifr_ifindex;
}


/*
 * Clear a certain interface flag.
 */
int
clr_flag (int *s, char *ifname, short flag)
{
  struct ifreq ifr;
  memset (&ifr, 0, sizeof (struct ifreq));
  strncpy (ifr.ifr_name, ifname, IFNAMSIZ);
  ifr.ifr_flags &= ~flag;
  if (ioctl (*s, SIOCSIFFLAGS, &ifr) < 0)
    {
      syslog (LOG_ERR, "clr_flag(%s,%d): SIOCSIFFLAGS: %s", ifname, flag,
	      strerror (errno));
      return -1;
    }
  return (0);
}


/*
 * Set a certain interface flag.
 */
int
set_flag (int *s, char *ifname, short flag)
{
  struct ifreq ifr;
  memset (&ifr, 0, sizeof (struct ifreq));
  strncpy (ifr.ifr_name, ifname, IFNAMSIZ);
  ifr.ifr_flags |= flag;
  if (ioctl (*s, SIOCSIFFLAGS, &ifr) < 0)
    {
      syslog (LOG_ERR, "set_flag(%s,%d): SIOCSIFFLAGS: %s", ifname, flag,
	      strerror (errno));
      return -1;
    }
  return (0);
}


/*
 * Set hw address
 */
  int
set_hwaddr (int *s, char *ifname, unsigned char *hwaddr)
{
  struct ifreq ifr;
  struct rtentry *r;
  int i = 0;

  r = malloc (sizeof (struct rtentry));
  memset (&ifr, 0, sizeof (struct ifreq));
  strncpy (ifr.ifr_name, ifname, IFNAMSIZ);

  /* get default gw for ifname if exist */
  if ((i = get_dg (r, ifname)) < 0)
    {
      free (r);
      r = NULL;
    }
  				/* if down */
    if (clr_flag (s, ifname, IFF_UP))
    {
      syslog (LOG_ERR, "set_hwaddr(): Shutting down if %s failed!\n", ifname);
      return (-1);
    }
  				/* get hw addr to populate struct */
    if (ioctl (*s, SIOCGIFHWADDR, &ifr) < 0)
    {
      syslog (LOG_ERR, "set_hwaddr(): SIOCGIFHWADDR: %s\n", strerror (errno));
      return (-1);
    }

  /* copy relevant data */
  memcpy (&ifr.ifr_hwaddr.sa_data, hwaddr, ETH_ALEN);
  				/* set hw addr */
    if (ioctl (*s, SIOCSIFHWADDR, &ifr) < 0)
    {
      syslog (LOG_ERR, "set_hwaddr(): SIOCSIFHWADDR: %s\n", strerror (errno));
      return (-1);
    }
  				/* if up */
    if (set_flag (s, ifname, IFF_UP))
    {
      syslog (LOG_ERR, "set_hwaddr(): Getting up if %s failed!\n", ifname);
      return (-1);
    }

  /* reset default gw after if shutdown */
  if (r != NULL)
    {
      if (ioctl (*s, SIOCADDRT, r) < 0)
	{
	  syslog (LOG_ERR, "vgp_change_status: SIOCADDRT: %s",
		  strerror (errno));
	  return (-1);
	}
    }

  if (debug)
    syslog (LOG_DEBUG,
	    "set_hwaddr(): Successfully set MAC address: %02X:%02X:%02X:%02X:%02X:%02X\n",
	    hwaddr[0], hwaddr[1], hwaddr[2], hwaddr[3], hwaddr[4], hwaddr[5]);
  return (0);
}


/*
 * return MAC address in hwaddr
 */
int
get_hwaddr (int *s, char *ifname, unsigned char *hwaddr)
{
  struct ifreq ifr;
  strncpy (ifr.ifr_name, ifname, IFNAMSIZ);

  if (ioctl (*s, SIOCGIFHWADDR, &ifr) == -1)
    {
      syslog (LOG_ERR, "get_hwaddr(%s): SIOCGIFHWADDR: %s", ifname,
	      strerror (errno));
      return (-1);
    }
  memcpy (hwaddr, &ifr.ifr_hwaddr.sa_data, ETH_ALEN);
  if (debug)
    syslog (LOG_DEBUG,
	    "get_hwaddr(): Successfully got our MAC address: %02X:%02X:%02X:%02X:%02X:%02X\n",
	    hwaddr[0], hwaddr[1], hwaddr[2], hwaddr[3], hwaddr[4], hwaddr[5]);
  return 0;
}


/*
 * set an ip address and other interface parameters
 * Inspired from http://www.sophics.cz/cgi-bin/cvsweb/aranym/src/natfeat/ethernet.cpp?rev=1.3
 */
int
set_ipaddr (int sockfd, char *ifname, int oper, char *opname, char *ipaddr)
{
    int             rc;                     /* Return code               */
    struct ifreq    ifreq;                  /* Structure for ioctl       */
    struct sockaddr_in *sin;                /* -> sockaddr within ifreq  */

    /* Initialize the structure for ioctl */
    memset (&ifreq, 0, sizeof(ifreq));
    strcpy (ifreq.ifr_name, ifname);

    /* Determine where in the structure to store the address */
    sin = (struct sockaddr_in*)
            (oper == SIOCSIFADDR ? &ifreq.ifr_addr
            :oper == SIOCSIFDSTADDR ? &ifreq.ifr_dstaddr
            :oper == SIOCSIFNETMASK ? &ifreq.ifr_netmask
            :oper == SIOCSIFBRDADDR ? &ifreq.ifr_broadaddr
            :NULL);

    /* Store the IP address into the structure */
    sin->sin_family = AF_INET;
    rc = inet_aton (ipaddr, &sin->sin_addr);
    if (rc == 0)
    {
      syslog (LOG_ERR, "Invalid %s (%s) for %s: %s\n", opname, ipaddr, ifname, strerror(errno));
      return -1;
    }

    /* Configure the interface with the specified IP address */
    rc = ioctl (sockfd, oper, &ifreq);

    if (rc < 0)
	{
	  syslog (LOG_ERR, "Error setting %s (%s) for %s: %s\n", opname, ipaddr, ifname, strerror(errno));
	  return -1;
	}

    return 0;
} /* end function set_ipaddr */



int
get_dg (struct rtentry *r, char *ifname)
{
  char iface[IFNAMSIZ];
  char temp[3];
  char *buf;
  int i = 0;
  FILE *f;

  unsigned char net_addr[128];
  unsigned char gate_addr[128];
  unsigned char mask_addr[128];
  int refcnt, use, metric, mss, window, irtt;
  unsigned char iflags;

  char *fmt = "%16s %128s %128s %X %d %d %d %128s %d %d %d\n";
  buf = malloc (512);

  if ((f = fopen ("/proc/net/route", "r")) == NULL)
    {
      syslog (LOG_ERR, "get_dg(): fopen: /proc/net/route: %s",
	      strerror (errno));
      return -1;
    }

  while (fgets (buf, 511, f) != NULL)
    {

      sscanf (buf, fmt, iface, net_addr, gate_addr,
	      &iflags, &refcnt, &use, &metric, mask_addr,
	      &mss, &window, &irtt);

      if ((strcmp (iface, ifname) == 0) && (iflags == 0x03))
	{			// 0x03 == UG
	  memset (r, 0, sizeof (struct rtentry));

	  r->rt_metric = 0;
	  r->rt_mtu = 40;
	  r->rt_flags = iflags;
	  r->rt_irtt = irtt;
	  r->rt_window = window;
	  r->rt_dst.sa_family = AF_INET;
	  r->rt_gateway.sa_family = AF_INET;
	  r->rt_genmask.sa_family = AF_INET;

	  while (i < 4)
	    {
	      strncpy (temp, &gate_addr[i * 2], 2);

	      //printf("%c\n", strtol(temp, NULL, 16));
	      /* the most significant ipaddress byte is on the right */
	      r->rt_gateway.sa_data[5 - i] = strtol (temp, NULL, 16);
	      i++;
	    }
	}
    }

  if (i == 4)
    return 0;
  else
    return -1;
}
