/*
     This file is part of vgp daemon.
     (C) 2003 Marco Tizzoni aka elibus (elibus@freaknet.org)
     Key available on key server.
     Key fingerprint: 3A97 6C78 0D07 91FC 0509  0661 DAAE F7B4 FC6F E313

     vgp daemon is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     vgpd is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with libextractor; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.

*/

#define SYSLOG_NAMES

#include <sys/socket.h>
#include <features.h>		/* for the glibc version number */
#if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 1
#include <netpacket/packet.h>
#include <net/ethernet.h>	/* the L2 protocols */
#else
#include <asm/types.h>
#include <linux/if_packet.h>
#include <linux/if_ether.h>	/* The L2 protocols */
#endif

#include <sys/ioctl.h>

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <syslog.h>
#include <errno.h>

# include <linux/route.h>

#include "vgp.h"
#include "if_util.h"
#include "sha1.h"
#include "cmdline.h"
#include "config.h"

#define TOT_LEN   ETH_HLEN+VGP_LEN
#define BUF_SIZE  TOT_LEN


int s = 0;			/* Socketdescriptor */
int debug = 0;
struct sockaddr_ll sa;

/* send and recv buffer */
unsigned char *send_buffer;
unsigned char *recv_buffer;
unsigned char *temp_buffer;

/* temp mac */
unsigned char real_mac[6];	/* our MAC address */
unsigned char dest_mac[6] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };	/* broadcast! */

/*********************/
/*                   */
/* vgp configuration */
/*                   */
/*********************/
unsigned short int status = STANDBY;

/* mandatory vars  */
unsigned char vmac[6] = { 0x00, 0x00, 0x5E, 0x00, 0x01, 0x01 };
char *vip = "undefined"; /* Virtual IP address of machine */

/* customizable vars */
char ifname[IFNAMSIZ] = "dummy0";	/* listen on... */
char vifname[IFNAMSIZ] = "dummy0:1";	/* MUST BE THE SAME FISICAL IF */
unsigned char mypriority = 0xFF;	/* default max priority */
unsigned char mygroup = 0x01;	/* default group 1 */
unsigned char hello_time = 2;	/* default hello_time */
unsigned char hold_time = 6;	/* default hold_time */
char *netmask = "using default netmask"; /* (optional) netmask from command line */
int netmask_given = 0;
char *broadcast = "using default broadcast address"; /* (optional) bcast addr from command line */
int broadcast_given = 0;


unsigned char key[KEY_LEN];

int logto = LOG_DAEMON;		/* default log to DAEMON */


void
cleanup (int signum)
{

  /* Clean up....... */
  if (debug)
    syslog (LOG_DEBUG, "cleanup()");
  if (s == -1)
    return;

  /* shutdown alias if */
  if (status == ACTIVE)
    {
      if (clr_flag (&s, vifname, IFF_UP) < 0)
	{
	  syslog (LOG_ERR, "vgp_change_status(): shutting down!\n");
	  exit (-1);
	}
    }

  /* reset real_mac */
  if (set_hwaddr (&s, ifname, real_mac) < 0)
    {
      syslog (LOG_ERR, "vgp_change_status(): Setting hwaddr failed!\n");
      exit (-1);
    }

  close (s);

  free (send_buffer);
  free (recv_buffer);
  free (temp_buffer);

  syslog (LOG_INFO, "Server terminating....\n");
  closelog ();
  exit (0);

}

void
vgp_change_status (unsigned short int nexts)
{
  int i = 0;

  if (debug)
    syslog (LOG_DEBUG, "vgp_change_status(): status == %d", nexts);

  if (nexts == ACTIVE)
    {
      status = ACTIVE;

      if (set_hwaddr (&s, ifname, vmac) < 0)
	{			// set virtual hw addr
	  syslog (LOG_ERR, "vgp_change_status(): Setting hwaddr failed!\n");
	  exit (-1);
	}

      if (set_ipaddr (s, vifname, SIOCSIFADDR, "virtual interface IP address", vip) < 0)
	{
	  syslog (LOG_ERR, "vgp_change_status(): Setting ipaddr failed!\n");
	  exit (-1);
	}

	if (netmask_given)
		if (set_ipaddr (s, vifname, SIOCSIFNETMASK, "virtual interface netmask", netmask) < 0)
		{
			syslog (LOG_ERR, "vgp_change_status(): Setting netmask failed!\n");
			exit (-1);
		}

	if (broadcast_given)
		if (set_ipaddr (s, vifname, SIOCSIFBRDADDR, "virtual interface broadcast address", broadcast) < 0)
		{
			syslog (LOG_ERR, "vgp_change_status(): Setting broadcast address failed!\n");
			exit (-1);
		}

	syslog (LOG_INFO, "Status changed: ACTIVE\n");
	alarm (hello_time);

    }
  else if (nexts == STANDBY)
    {
      status = STANDBY;
      if (set_hwaddr (&s, ifname, real_mac) < 0)
	{			// reset real hw addr
	  syslog (LOG_ERR, "vgp_change_status(): Setting hwaddr failed!\n");
	  exit (-1);
	}
      if (clr_flag (&s, vifname, IFF_UP) < 0)
	{
	  syslog (LOG_ERR, "vgp_change_status(): shutting down if %s!\n",
		  vifname);
	  exit (-1);
	}
      syslog (LOG_INFO, "Status changed: STANDBY\n");
      alarm (hold_time);

    }
  else
    {
      syslog (LOG_ERR, "Status not handled! :-( \n");
      syslog (LOG_ERR, "Please send an e-mail to elibus@freaknet.org\n");
      exit (-1);
    }

}


void
vgp_hello_init ()
{
  int i;

  /* Pointer used by send_hello */
  struct ethhdr *s_eh;
  struct vgp *s_vgp;
  HashCode160 hc;
  unsigned char temp[ETH_HLEN + VGP_HLEN + KEY_LEN];

  s_eh = (struct ethhdr *) send_buffer;	/* Another pointer to ethernet header */
  s_vgp = (struct vgp *) (send_buffer + ETH_HLEN);	/* pointer to vgp */

  if (debug)
    syslog (LOG_DEBUG, "vgp_hello_init()");

  /*prepare sockaddr_ll */
  sa.sll_family = PF_PACKET;
  sa.sll_protocol = htons (ETH_P_IP);
  sa.sll_ifindex = get_ifindex (&s, vifname);
  sa.sll_hatype = 0;
  sa.sll_pkttype = PACKET_BROADCAST;
  sa.sll_halen = ETH_ALEN;
  for (i = 0; i < ETH_ALEN; i++)
    sa.sll_addr[i] = 0xFF;
  /* not used */
  sa.sll_addr[6] = 0x00;
  sa.sll_addr[7] = 0x00;

  /* fill ethernet address field */
  memcpy ((void *) send_buffer, (void *) dest_mac, ETH_ALEN);
  memcpy ((void *) (send_buffer + ETH_ALEN), (void *) real_mac, ETH_ALEN);

  s_eh->h_proto = ETH_P_802_3;
  s_vgp->priority = mypriority;
  s_vgp->group = mygroup;

  memset (&hc, 0, sizeof (HashCode160));	// reset struct

  /* copy header and key in a temp buffer */
  memcpy (temp, s_eh, ETH_HLEN + VGP_HLEN);
  memcpy (temp + ETH_HLEN + VGP_HLEN, key, KEY_LEN);

  /* hash and take result */
  hash (temp, BUF_SIZE, &hc);
  memcpy (&s_vgp->key, &hc, VGP_KLEN);

}

int
vgp_hello_send ()
{
  if (debug)
    syslog (LOG_DEBUG, "vgp_hello_send()");

  if (sendto (s, send_buffer, TOT_LEN, 0,
	      (struct sockaddr *) &sa, sizeof (sa)) == -1)
    {
      syslog (LOG_ERR, "sendto(): %s", strerror (errno));
      if (errno == ENETDOWN)	/* Newtork may be down while setting */
	return 0;		/* VMAC so return no error */
      else
	return (-1);
    }

  if (debug)
    syslog (LOG_DEBUG, "vgp_hello_send(): frame sent!\n");

  return 0;
}


void
vgp_timer ()
{
  if (debug)
    syslog (LOG_DEBUG, "vgp_timer()");

  if (status == ACTIVE)
    {
      vgp_hello_send ();
      alarm (hello_time);
    }
  else if (status == STANDBY)
    {
      vgp_change_status (ACTIVE);
    }
  else
    {
      syslog (LOG_ERR, "Status not handled! :-( \n");
      syslog (LOG_ERR, "Please send an e-mail to elibus@freaknet.org\n");
      exit (-1);
    }

}


int
usage ()
{

  printf ("Usage: vgpd [OPTIONS] <-i ip_address> \n");
  printf ("\t-a ip address to configure on virtual interface\n");
  printf ("\n");
  printf ("OPTIONS:\n");
  printf ("\t-m Virtual router ID  (default 01)\n");
  printf ("\t-i virtual interface  (default dummy0:1)\n");
  printf ("\t-h vgp hello time     (default 2sec)\n");
  printf ("\t-p vgp priority       (default 255)\n");
  printf ("\t-g vgp group          (default 1)\n");
  printf ("\t-l log facility       (default daemon)\n");
  printf ("\t-d debug\n");
  exit (-1);

}

void
print_param ()
{

  printf ("Interface           : %s\n", ifname);
  printf ("Virtual interface   : %s\n", vifname);
  printf ("vgp priority set to : 0x%02X\n", mypriority);
  printf ("vgp group set to    : 0x%02X\n", mygroup);
  printf ("Hello time          : %d\n", hello_time);
  printf ("Hold time           : %d\n", hold_time);
  printf ("Virtual IP          : %s\n", vip);
  printf ("Virtual Netmask     : %s\n", netmask);
  printf ("Virtual Broadcast   : %s\n", broadcast);
  printf ("Virtual MAC         : %02X:%02X:%02X:%02X:%02X:%02X\n",
	  vmac[0], vmac[1], vmac[2], vmac[3], vmac[4], vmac[5]);
  syslog (LOG_INFO, "Interface           : %s\n", ifname);
  syslog (LOG_INFO, "Virtual interface   : %s\n", vifname);
  syslog (LOG_INFO, "vgp priority set to : 0x%02X\n", mypriority);
  syslog (LOG_INFO, "vgp group set to    : 0x%02X\n", mygroup);
  syslog (LOG_INFO, "Hello time          : %d\n", hello_time);
  syslog (LOG_INFO, "Hold time           : %d\n", hold_time);
  syslog (LOG_INFO, "Virtual IP          : %s\n", vip);
  syslog (LOG_INFO, "Virtual Netmask     : %s\n", netmask);
  syslog (LOG_INFO, "Virtual Broadcast   : %s\n", broadcast);
  syslog (LOG_INFO,
	  "Virtual MAC         : %02X:%02X:%02X:%02X:%02X:%02X\n",
	  vmac[0], vmac[1], vmac[2], vmac[3], vmac[4], vmac[5]);

}

void
read_key ()
{
  FILE *keyfile;
  int i = 0;
  char c;

  memset (&key, 0, KEY_LEN);

  if ((keyfile = fopen (KEYFILE, "r")) == NULL)
    {
      syslog (LOG_ERR, "fopen(): %s: %s\n", KEYFILE, strerror (errno));
      exit (-1);
    }

  while (((c = getc (keyfile)) != EOF) && (i < KEY_LEN))
    {
      key[i] = c;
      i++;
    }

}


int
main (int argc, char **argv)
{
  char c;
  short i = 0;

  int length;			/* length of received packet */
  struct ethhdr *r_eh;
  struct vgp *r_vgp, *t_vgp;
  HashCode160 hc;
  unsigned char temp[ETH_HLEN + VGP_HLEN + KEY_LEN];	/* to pass to hash() */

  static struct gengetopt_args_info args_info;

  struct sockaddr_ll sall;


  /* Allocate necessary memory */
  send_buffer = malloc (BUF_SIZE);
  recv_buffer = malloc (BUF_SIZE);
  temp_buffer = malloc (BUF_SIZE);
  r_eh = (struct ethhdr *) recv_buffer;	/* Another pointer to ethernet header */
  r_vgp = (struct vgp *) (recv_buffer + ETH_HLEN);	/* pointer to vgp */
  t_vgp = (struct vgp *) (temp_buffer + ETH_HLEN);	/* pointer to vgp */


  /* Parse command line
   ** v = virtual if, h = hello_time, H = hold_time, l = log_facility,
   ** p = priority, g = group, i = virtual_ip, m = virtual_mac  */

  if (cmdline_parser (argc, argv, &args_info) != 0)
    exit (1);

  if (args_info.help_given)
    cmdline_parser_print_help ();

  if (args_info.version_given)
    cmdline_parser_print_version ();

  if (args_info.debug_given)
    debug = 1;

  if (args_info.virtual_ip_given) vip = args_info.virtual_ip_arg;

  if (args_info.virtual_netmask_given) {
    netmask = args_info.virtual_netmask_arg;
    netmask_given = 1;
  }

  if (args_info.virtual_broadcast_given) {
    broadcast = args_info.virtual_broadcast_arg;
    broadcast_given = 1;
  }

  if (args_info.virtual_if_given)
    {
      /* reset */
      memset (ifname, '\0', IFNAMSIZ);
      memset (vifname, '\0', IFNAMSIZ);
      strncpy (vifname, args_info.virtual_if_arg, IFNAMSIZ - 1);
      i = 0;
      while ((vifname[i] != '\0') && (vifname[i] != ':')
	     && (i < IFNAMSIZ - 1))
	{
	  ifname[i] = vifname[i];
	  i++;
	}
      ifname[i] = '\0';

    }

  if (args_info.group_given)
    {
      mygroup = args_info.group_arg;
      vmac[5] = mygroup;

    }

  if (args_info.priority_given)
    mypriority = args_info.priority_arg;

  if (args_info.log_given)
    {
      i = 0;
      while (facilitynames[i].c_val != -1)
	{
	  if (strcmp (facilitynames[i].c_name, args_info.log_arg) == 0)
	    {
	      logto = facilitynames[i].c_val;
	      printf ("Logging to          : %s\n", facilitynames[i].c_name);
	      i = -1;
	      break;
	    }
	  i++;
	}
      if (i != -1)
	{
	  fprintf (stderr, "Are you sure you want to log to: %s???\n",
		   optarg);
	  exit (-1);
	}

    }

  if (args_info.hello_time_given)
    {
      if ((hello_time = args_info.hello_time_arg) == 0)
	{
	  fprintf (stderr,
		   "*** error: Hello time set to %d sec is NOT a good idea!\n",
		   hello_time);
	  exit (-1);
	}
      if (hello_time > 5)
	fprintf (stderr,
		 "*** warning: Hello time set to %d sec is NOT a good idea!\n",
		 hello_time);
      hold_time = hello_time * 3;

    }


  /* logging stuff */
  openlog (argv[0], LOG_PID | LOG_CONS, logto);
  syslog (LOG_INFO, "vgpd started, entering initialization phase...\n");
  print_param ();

  if (daemon (0, 0) < 0)
    fprintf (stderr, "Cannot daemonize!\n");

  read_key ();

  /* open socket */
  s = socket (AF_PACKET, SOCK_RAW, htons (ETH_P_ALL));
  if (s == -1)
    {
      syslog (LOG_ERR, "socket(): %s\n", strerror (errno));
      exit (-1);
    }

  if (debug)
    syslog (LOG_INFO, "Successfully opened socket: %i\n", s);

  /* bind to specific interface */
  memset (&sall, 0, sizeof (struct sockaddr_ll));

  sall.sll_family = PF_PACKET;
  sall.sll_protocol = htons (ETH_P_ALL);
  sall.sll_ifindex = get_ifindex (&s, ifname);

  if (bind (s, &sall, sizeof (struct sockaddr_ll)) < 0)
    {
      perror ("bind()");
      exit (-1);
    }

  if (debug)
    syslog (LOG_INFO, "Successfully opened socket: %i\n", s);

  if (get_hwaddr (&s, ifname, real_mac) < 0)
    {
      syslog (LOG_ERR, "get_hwaddr(): %s\n", strerror (errno));
      exit (-1);
    };

  /* establish signal handler */
  signal (SIGINT, cleanup);	/* clean up */
  signal (SIGTERM, cleanup);
  signal (SIGALRM, vgp_timer);	/* handle timer and status */

  vgp_hello_init ();		/* create the frames */

  alarm (hello_time + 1);

  while (1)
    {

      /* Wait for incoming packet... (non blocking) */
      length = recvfrom (s, recv_buffer, BUF_SIZE, 0, NULL, NULL);
      if (length == -1)
	{
	  if (errno == ENETDOWN);
	  else
	    {
	      syslog (LOG_ERR, "recvfrom(): %s", strerror (errno));
	      exit (-1);
	    }

	}
      else if (length == TOT_LEN)
	{			/* VGP frames have fixed length */
	  if (debug)
	    {
	      syslog (LOG_DEBUG, "Packet received!");
	      syslog (LOG_DEBUG,
		      "PROTO: %02X, VGP_GROUP: %02X, VGP_PRIORITY: %02X \n",
		      r_eh->h_proto, r_vgp->priority, r_vgp->group);
	    }

	  /* See if we should process */
	  /* Group check */
	  if (debug)
	    syslog (LOG_DEBUG, "check proto...OK!\n");
	  if (r_vgp->group == mygroup)
	    {
	      if (debug)
		syslog (LOG_DEBUG, "check group...OK!\n");

	      /* Priority check */
	      if (r_vgp->priority > mypriority)
		{		/* Ops...there's a master on the net */

		  /* hash verify */
		  memset (&hc, 0, sizeof (HashCode160));
		  memcpy (temp, r_eh, ETH_HLEN + VGP_HLEN);	/* ready for hash */
		  memcpy (temp + ETH_HLEN + VGP_HLEN, key, KEY_LEN);
		  hash (temp, BUF_SIZE, &hc);

		  /* starts key check */
		  i = 0;
		  while (i < VGP_KLEN)
		    {
		      if (hc.data[i] != r_vgp->key[i])
			{
			  syslog (LOG_INFO, "Authentication failure!\n");
			  break;
			}
		      i++;
		    }
		  if (i == VGP_KLEN)
		    {
		      if (debug)
			{
			  syslog (LOG_DEBUG,
				  "Found a vgp with higher priority %02X\n",
				  r_vgp->priority);
			  syslog (LOG_DEBUG,
				  "Setting hold_time alarm to: %i\n",
				  hold_time);
			}
		      alarm (hold_time);
		      if (status == ACTIVE)
			vgp_change_status (STANDBY);
		    }
		}
	    }
	}
    }
}
